from django.db import models


class UserModel(models.Model):
    """
        BaseModel to store the user_id and user_email of the user that is logged in the service.
        Typically used to replace the Django User model instance.
    """

    user_email = models.EmailField(null=True)
    user_auth_id = models.PositiveIntegerField(null=True)

    class Meta:
        abstract = True



from django.conf import settings

from rest_framework.exceptions import APIException
from rest_framework.generics import ListAPIView
from rest_framework.views import APIView

from .authenticators import JWTAuthentication
from .permissions import HasBackofficePermission


class ConfigurationError(APIException):
    status_code = 500  # Internal Server Error
    default_detail = "Misconfigured view: cannot determine permission action."
    default_code = 'misconfigured_view'


class BasePermissionMixin:
    authentication_classes = [JWTAuthentication]
    permission_classes = [HasBackofficePermission]

    permission_page = None
    permission_action_list = None
    permission_action_retrieve = None
    permission_action_create = None
    permission_action_patch = None
    permission_action_put = None
    permission_action_delete = None

    def set_permissions_from_mapping(self, parameter_value, mappings):
        for mapping in mappings:
            if mapping.get('value') == parameter_value:
                for attr, value in mapping.items():
                    if attr != 'value':
                        setattr(self, attr, value)
                break

    def get_permissions(self):
        if settings.MONI_AUTH["JWT_DEBUG"]:
            return super().get_permissions()

        action_permission_mapping = {
            'list': 'permission_action_list',
            'retrieve': 'permission_action_retrieve',
            'create': 'permission_action_create',
            'partial_update': 'permission_action_patch',
            'patch': 'permission_action_patch',
            'put': 'permission_action_patch',
            'update': 'permission_action_put',
            'delete': 'permission_action_delete',
            'get': 'permission_action_list',
            'post': 'permission_action_create',
        }

        # Check if 'action' attribute is available (for ViewSets or ModelViewSets)
        if hasattr(self, 'action'):
            action = self.action

            # Check for custom action permissions
            if action not in action_permission_mapping and hasattr(self, 'custom_action_permissions'):
                custom_permissions = getattr(
                    self, 'custom_action_permissions', {})

                # If a custom action is defined, directly set the permission_action
                if action in custom_permissions:
                    self.permission_action = custom_permissions[action]
                    return super().get_permissions()

        # Check if the view is an instance of APIView or ListAPIView
        elif isinstance(self, (APIView, ListAPIView)):
            action = self.request.method.lower()
        else:
            raise ConfigurationError(
                "Unable to determine action for permission check")

        permission_attribute = action_permission_mapping.get(action)
        if permission_attribute is not None and getattr(self, permission_attribute, None) is not None:
            self.permission_action = getattr(self, permission_attribute)
        else:
            raise ConfigurationError(
                f"No permission attribute mapping found for action: {action}")

        return super().get_permissions()


class ParameterBasedPermissionMixin(BasePermissionMixin):
    """
    Mixin to handle permission logic based on a parameter value.
    """
    parameter_name = None
    parameter_permissions_map = []

    def get_permissions(self):
        parameter_value = self.kwargs.get(self.parameter_name)
        self.set_permissions_from_mapping(
            parameter_value, self.parameter_permissions_map)
        return super().get_permissions()


class DataParameterBasedPermissionMixin(BasePermissionMixin):
    """
    Mixin to handle permission logic based on a parameter value in request data.
    """
    data_parameter_name = None
    data_parameter_permissions_map = []

    def get_permissions(self):
        parameter_value = self.request.data.get(self.data_parameter_name)
        self.set_permissions_from_mapping(
            parameter_value, self.data_parameter_permissions_map)
        return super().get_permissions()



from rest_framework.exceptions import AuthenticationFailed, PermissionDenied
from rest_framework.status import HTTP_401_UNAUTHORIZED
from rest_framework.views import exception_handler


def custom_exception_handler(exc, context):
    """
        Exception handler that serves two purposes:
        - Returns HTTP 401 Unauthorized when the AuthenticationFailed exception is raised.
        This is commonly raised when the JWT is invalid or expired in the JWTAuthentication class.
        - Modifies the message when a PermissionDenied is raised so it has a key-structure format.
    """
    # Call the default exception handler first
    response = exception_handler(exc, context)

    if isinstance(exc, AuthenticationFailed):
        # Modify the response status code to 401 Unauthorized
        response.status_code = HTTP_401_UNAUTHORIZED
    elif isinstance(exc, PermissionDenied):
        # Modify the response data to add a key-structure
        response.data = {"detail": PermissionDenied.default_code}

    return response



import jwt

from http.cookies import SimpleCookie

from django.conf import settings


from rest_framework.test import APITestCase
from unittest.mock import patch

from .utils import get_pages_keys, get_products_keys, get_actions_keys


class BackofficeAuthenticatedTestCase(APITestCase):
    def setUp(self):
        self._mock_signals()

        payload = {
            "sub": 1,
            "email": "test@moni.com.ar",
            "permissions": {
                "products": get_products_keys(),
                "pages": get_pages_keys(),
                "actions": get_actions_keys(),
            }
        }
        token = jwt.encode(
            payload, settings.MONI_AUTH["JWT_SECRET_KEY"], algorithm="HS256")

        self.client.cookies = SimpleCookie()
        for key in settings.MONI_AUTH["JWT_KEYS"]:
            self.client.cookies[key] = token

    def _mock_signals(self):
        # Mock the signal handlers to do nothing
        patcher1 = patch(
            'moni_auth.signals.django_admin_user_action', lambda *args, **kwargs: None)
        patcher2 = patch(
            'moni_auth.signals.django_admin_user_login', lambda *args, **kwargs: None)
        patcher3 = patch(
            'moni_auth.signals.django_admin_user_logout', lambda *args, **kwargs: None)
        self.addCleanup(patcher1.stop)
        self.addCleanup(patcher2.stop)
        self.addCleanup(patcher3.stop)
        patcher1.start()
        patcher2.start()
        patcher3.start()



from django.conf import settings
from django.contrib.admin.models import LogEntry
from django.contrib.auth import user_logged_in, user_logged_out
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone

from .service import AuthenticationService


@receiver(post_save, sender=LogEntry)
def django_admin_user_action(sender, instance, **kwargs):
    if settings.ENVIRONMENT == "test":
        return

    # Check if the change was made in the admin site
    if instance.is_change():
        action = 'UPDATED'
    elif instance.is_addition():
        action = 'CREATED'
    elif instance.is_deletion():
        action = 'DELETED'
    else:
        return

    details = {
        "email": instance.user.email,
        "instance": {
            "id": instance.object_id,
            "name": instance.content_type.model,
            "object_repr": instance.object_repr,
        },
        "type": action,
        "message": f"{instance.user.email} {action} an instance of {instance.content_type.model} with ID {instance.object_id}",  # noqa
        "service": settings.MONI_AUTH["SERVICE_NAME"],
        "action_time": instance.action_time,
    }
    AuthenticationService().send_log(details)


@receiver(user_logged_in)
def django_admin_user_login(sender, request, user, **kwargs):
    if settings.ENVIRONMENT == "test":
        return

    action = "LOGGED IN"
    details = {
        "email": user.email,
        "type": action,
        "message": f"{user.email} has {action} into the {settings.MONI_AUTH['SERVICE_NAME']} service",
        "service": settings.MONI_AUTH["SERVICE_NAME"],
        "action_time": timezone.now()
    }
    AuthenticationService().send_log(details)


@receiver(user_logged_out)
def django_admin_user_logout(sender, request, user, **kwargs):
    if settings.ENVIRONMENT == "test":
        return

    action = "LOGGED OUT"
    details = {
        "email": user.email,
        "type": action,
        "message": f"{user.email} has {action} from the {settings.MONI_AUTH['SERVICE_NAME']} service",
        "service": settings.MONI_AUTH["SERVICE_NAME"],
        "action_time": timezone.now(),
    }
    AuthenticationService().send_log(details)



from django.conf import settings

from rest_framework.permissions import BasePermission

from .authenticators import decode_jwt


class BaseJWTPermission(BasePermission):
    product = None
    page = None
    action = None

    def has_permission(self, request, view):
        if settings.MONI_AUTH["JWT_DEBUG"]:
            return True

        permissions = {'products': set(), 'pages': set(), 'actions': set()}

        for jwt_key in settings.MONI_AUTH["JWT_KEYS"]:
            decoded_token = decode_jwt(request.COOKIES.get(jwt_key))
            for perm_key, perm_values in decoded_token.get("permissions", {}).items():
                permissions[perm_key].update(perm_values)

        permissions = {k: list(v) for k, v in permissions.items()}

        self.page = getattr(view, "permission_page", None)
        self.action = getattr(view, "permission_action", None)

        if self.product and self.product not in permissions.get("products", []):
            return False

        if self.page and self.page not in permissions.get("pages", []):
            return False

        if self.action and self.action not in permissions.get("actions", []):
            return False

        return True


class HasBackofficePermission(BaseJWTPermission):
    product = settings.MONI_AUTH["PRODUCTS"]["backoffice"]


class HasDjangoAdminPermission(BaseJWTPermission):
    product = settings.MONI_AUTH["PRODUCTS"]["admin_de_django"]


class HasSharedSecretPermission(BasePermission):
    """
    Custom permission to check for a shared secret token in the request header.
    """

    def has_permission(self, request, view):
        secret_token = request.headers.get('X-Shared-Secret')
        return secret_token == settings.MONI_AUTH["SHARED_SECRET_TOKEN"]
